# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import netaddr
import six

from infoblox_client import objects
from networking_infoblox.neutron.common import constants as const


@six.add_metaclass(abc.ABCMeta)
class IPAllocator(object):
    """Base class for ip allocation actions"""

    DEFAULT_OPTIONS = {
        'use_host_record': True,
        'configure_for_dhcp': True,
        'configure_for_dns': True,
        'dns_record_binding_types': [],
        'dns_record_unbinding_types': [],
        'dns_record_removable_types': []}

    def __new__(cls, ib_obj_manager, options):
        cls._validate_and_set_default_options(options)
        return super(IPAllocator, cls).__new__(
            cls._get_ip_allocator_class(options))

    def __init__(self, ib_obj_manager, options):
        self.manager = ib_obj_manager
        self.opts = options

    @classmethod
    def _validate_and_set_default_options(cls, options):
        if not isinstance(options, dict):
            raise ValueError('Options should be passed as dict')
        for key in cls.DEFAULT_OPTIONS:
            if key not in options:
                options[key] = cls.DEFAULT_OPTIONS[key]

    @staticmethod
    def _get_ip_allocator_class(options):
        if options['use_host_record']:
            return HostRecordIPAllocator
        else:
            return FixedAddressIPAllocator

    @abc.abstractmethod
    def allocate_ip_from_range(self, network_view, dns_view, zone_auth,
                               hostname, mac, first_ip, last_ip,
                               extattrs=None):
        pass

    @abc.abstractmethod
    def allocate_given_ip(self, network_view, dns_view, zone_auth,
                          hostname, mac, ip, extattrs=None):
        pass

    @abc.abstractmethod
    def deallocate_ip(self, network_view, dns_view_name, ip):
        pass

    @abc.abstractmethod
    def bind_names(self, network_view, dns_view, ip, name, extattrs):
        pass

    @abc.abstractmethod
    def unbind_names(self, network_view, dns_view, ip, name, extattrs):
        pass


class HostRecordIPAllocator(IPAllocator):

    def bind_names(self, network_view, dns_view, ip, name, extattrs):
        # Don't use network view for DNS hosts
        net_view = None
        if not self.opts['configure_for_dns']:
            dns_view = None
            net_view = network_view

        # See OPENSTACK-181. In case hostname already exists on NIOS, update
        # host record which contains that hostname with the new IP address
        # rather than creating a separate host record object
        reserved_hostname_hr = self.manager.find_hostname(
            dns_view, name, ip, net_view)
        reserved_ip_hr = self.manager.get_host_record(
            dns_view, ip, net_view)

        if (reserved_hostname_hr and reserved_ip_hr and
                reserved_hostname_hr.ref == reserved_ip_hr.ref):
            reserved_hostname_hr.extattrs = extattrs
            reserved_hostname_hr.update()
            return

        if reserved_hostname_hr:
            for hr_ip in reserved_ip_hr.ips:
                ip_mac = hr_ip.mac if hr_ip.ip_version == 4 else hr_ip.duid
                if hr_ip == ip:
                    reserved_ip_hr.delete()
                    reserved_hostname_hr.extattrs = extattrs
                    self.manager.add_ip_to_record(
                        reserved_hostname_hr, ip, ip_mac)
                    break
        else:
            self.manager.bind_name_with_host_record(
                dns_view, ip, name, extattrs, net_view)

    def unbind_names(self, network_view, dns_view, ip, name, extattrs):
        # Nothing to delete, all will be deleted together with host record.
        pass

    def allocate_ip_from_range(self, network_view, dns_view,
                               zone_auth, hostname, mac, first_ip, last_ip,
                               extattrs=None):
        use_dhcp = self.opts['configure_for_dhcp']
        use_dns = self.opts['configure_for_dns']
        fqdn = hostname + '.' + zone_auth
        host_record = self.manager.find_hostname(
            dns_view, fqdn, first_ip)

        if host_record:
            hr = self.manager.add_ip_to_host_record_from_range(
                host_record, network_view, mac, first_ip, last_ip, use_dhcp)
        else:
            # First search hosts with same MAC and if exists address within
            # given range - use it instead of creating new one
            # https://bugs.launchpad.net/networking-infoblox/+bug/1628517
            hosts = self.manager.find_host_records_by_mac(dns_view,
                                                          mac.lower(),
                                                          network_view)
            if hosts:
                ip_range = netaddr.IPRange(first_ip, last_ip)
                ip_version = netaddr.IPAddress(first_ip).version
                for host in hosts:
                    if host.ip_version != ip_version:
                        continue
                    for ip in host.ips:
                        ip_mac = ip.mac if ip_version == 4 else ip.duid
                        if ip_mac != mac:
                            continue
                        ip_addr = netaddr.IPAddress(ip.ip)
                        if ip_addr in ip_range:
                            self.manager.update_host_record_eas(dns_view,
                                                                ip.ip,
                                                                extattrs)
                            return ip.ip
            hr = self.manager.create_host_record_from_range(
                dns_view, network_view, zone_auth, hostname, mac,
                first_ip, last_ip, extattrs, use_dhcp, use_dns)
        return hr.ip[-1].ip

    def allocate_given_ip(self, network_view, dns_view, zone_auth,
                          hostname, mac, ip, extattrs=None):
        use_dhcp = self.opts['configure_for_dhcp']
        use_dns = self.opts['configure_for_dns']
        # First search hosts with same MAC and if exists address with
        # same IP - use it instead of creating new one
        # https://bugs.launchpad.net/networking-infoblox/+bug/1628517
        ip_obj = objects.IP.create(ip=ip, mac=mac)
        ip_version = ip_obj.ip_version
        ip_mac = ip_obj.mac if ip_version == 4 else ip_obj.duid

        hosts = self.manager.find_host_records_by_mac(dns_view,
                                                      ip_mac,
                                                      network_view)
        if hosts:
            for host in hosts:
                if host.ip_version != ip_version:
                    continue
                for host_ip in host.ips:
                    host_ip_mac = (
                        host_ip.mac if ip_version == 4 else host_ip.duid)
                    if host_ip_mac != ip_mac:
                        continue
                    if host_ip.ip == ip:
                        self.manager.update_host_record_eas(dns_view,
                                                            host_ip.ip,
                                                            extattrs)
                        return host_ip.ip
        hr = self.manager.create_host_record_for_given_ip(
            dns_view, zone_auth, hostname, mac, ip, extattrs, use_dhcp,
            use_dns)
        return hr.ip[-1].ip

    def deallocate_ip(self, network_view, dns_view_name, ip):
        host_record = self.manager.get_host_record(dns_view_name, ip)
        if host_record:
            if len(host_record.ip) > 1:
                self.manager.delete_ip_from_host_record(host_record, ip)
            else:
                host_record.delete()


class FixedAddressIPAllocator(IPAllocator):

    def bind_names(self, network_view, dns_view, ip, name, extattrs):
        bind_cfg = self.opts['dns_record_binding_types']
        device_owner = extattrs.get(const.EA_PORT_DEVICE_OWNER)
        if device_owner in const.NEUTRON_FLOATING_IP_DEVICE_OWNERS:
            self.manager.update_fixed_address_eas(network_view, ip, extattrs)
            if self.opts['configure_for_dns']:
                self.manager.update_dns_record_eas(dns_view, ip, extattrs)
        if bind_cfg and self.opts['configure_for_dns']:
            self.manager.bind_name_with_record_a(
                dns_view, ip, name, bind_cfg, extattrs)

    def unbind_names(self, network_view, dns_view, ip, name, extattrs):
        unbind_cfg = self.opts['dns_record_unbinding_types']
        if unbind_cfg and self.opts['configure_for_dns']:
            self.manager.unbind_name_from_record_a(
                dns_view, ip, name, unbind_cfg)

    def allocate_ip_from_range(self, network_view, dns_view,
                               zone_auth, hostname, mac, first_ip, last_ip,
                               extattrs=None):
        # First search addresses with same MAC and if exists address within
        # given range - use it instead of creating new one
        # https://bugs.launchpad.net/networking-infoblox/+bug/1628517
        fixed_addrs = self.manager.get_fixed_addresses_by_mac(network_view,
                                                              mac.lower())
        if fixed_addrs:
            ip_range = netaddr.IPRange(first_ip, last_ip)
            for fixed_addr in fixed_addrs:
                ip_addr = netaddr.IPAddress(fixed_addr.ip)
                if ip_addr in ip_range:
                    self.manager.update_fixed_address_eas(network_view,
                                                          fixed_addr.ip,
                                                          extattrs)
                    return fixed_addr.ip
        fa = self.manager.create_fixed_address_from_range(
            network_view, mac, first_ip, last_ip, extattrs)
        return fa.ip

    def allocate_given_ip(self, network_view, dns_view, zone_auth,
                          hostname, mac, ip, extattrs=None):
        # First search addresses with same MAC and if exists address with
        # same IP - use it instead of creating new one
        # https://bugs.launchpad.net/networking-infoblox/+bug/1628517
        fixed_addrs = self.manager.get_fixed_addresses_by_mac(network_view,
                                                              mac.lower())
        if fixed_addrs:
            for fixed_addr in fixed_addrs:
                if fixed_addr.ip == ip:
                    self.manager.update_fixed_address_eas(network_view, ip,
                                                          extattrs)
                    return fixed_addr.ip
        fa = self.manager.create_fixed_address_for_given_ip(
            network_view, mac, ip, extattrs)
        return fa.ip

    def deallocate_ip(self, network_view, dns_view_name, ip):
        delete_cfg = self.opts['dns_record_removable_types']
        if delete_cfg and self.opts['configure_for_dns']:
            self.manager.unbind_name_from_record_a(dns_view_name, ip,
                                                   None, delete_cfg)
        self.manager.delete_fixed_address(network_view, ip)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import netaddr

from infoblox_client import exceptions as ib_exc
from neutron.ipam import utils as ipam_utils
from neutron.plugins.ml2 import config as ml2_config
from neutron.tests.unit import testlib_api
from neutron_lib import context
from oslo_config import cfg

from networking_infoblox.ipam import driver as drv
from networking_infoblox.neutron.common import grid
from networking_infoblox.tests import base
from networking_infoblox.tests.unit import grid_sync_stub


class TestDriver(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):
        super(TestDriver, self).setUp()
        self.ctx = context.get_admin_context()
        self._setup_config()
        self.grid_stub = grid_sync_stub.GridSyncStub(self.ctx,
                                                     self.connector_fixture)
        self.grid_stub.prepare_grid_manager(wapi_version='2.0')
        self.grid_mgr = self.grid_stub.get_grid_manager()
        self.grid_mgr._report_sync_time = mock.Mock()
        self.grid_mgr.mapping._sync_nios_for_network_view = mock.Mock()
        self.grid_mgr.member._discover_dns_settings = mock.Mock(
            return_value={})
        self.grid_mgr.member._discover_dhcp_settings = mock.Mock(
            return_value={})
        self.grid_mgr.sync()
        self.grid_mgr.get_config = mock.Mock()

    def _setup_config(self):
        cfg.CONF.set_override('core_plugin',
                              'neutron.plugins.ml2.plugin.Ml2Plugin')
        ml2_config.cfg.CONF.set_override('type_drivers', 'local', group='ml2')

    def _mock_subnet(self, subnet_pool, requested_pools, cidr, gateway):
        ip_version = 4
        subnetpool_id = subnet_pool['id]'] if subnet_pool else None
        # if pools is passed, then allocation_pools should be a request format
        # which is IPRange type, but if not passed, then subnet format is
        # subnet dict.
        if requested_pools is None:
            requested_pools = ipam_utils.generate_pools(cidr, gateway)
            allocation_pools = [
                {'start': str(netaddr.IPAddress(p.first, ip_version)),
                 'end': str(netaddr.IPAddress(p.last, ip_version))}
                for p in requested_pools]
        else:
            allocation_pools = requested_pools

        return {
            'id': 'subnet-id',
            'cidr': cidr,
            'tenant_id': 'tenant-id',
            'gateway_ip': gateway,
            'name': 'subnet-name',
            'network_id': 'network-id',
            'subnetpool_id': subnetpool_id,
            'allocation_pools': allocation_pools,
            'ip_version': ip_version,
            'enable_dhcp': True}

    @mock.patch.object(grid, 'GridManager', mock.Mock())
    def _mock_driver(self, subnet_pool=None, requested_pools=None,
                     cidr='192.168.10.0/24', gateway='192.168.10.1'):
        driver = drv.InfobloxPool(subnet_pool, self.ctx)
        driver._grid_manager = self.grid_mgr

        driver._subnetpool = subnet_pool if subnet_pool else None
        subnet = self._mock_subnet(subnet_pool, requested_pools, cidr, gateway)
        driver._fetch_subnet = mock.Mock(return_value=subnet)
        return driver

    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_get_subnet(self, ib_cxt_mock, ipam_mock):
        driver = self._mock_driver()
        subnet_id = 'subnet-id'
        ipam_subnet = driver.get_subnet(subnet_id)

        ipam_mock.IpamSyncController.get_subnet.called_once_with()
        self.assertIsInstance(ipam_subnet, drv.InfobloxSubnet)

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_allocate_subnet(self, ib_cxt_mock, ipam_mock, dns_mock):
        pools = [netaddr.ip.IPRange('192.168.10.3', '192.168.10.25')]
        driver = self._mock_driver(requested_pools=pools)
        subnet_factory = driver.get_subnet_request_factory()
        subnet_request = subnet_factory.get_request(self.ctx,
                                                    driver._fetch_subnet(),
                                                    None)

        ipam_subnet = driver.allocate_subnet(subnet_request)

        ipam_mock.IpamSyncController.create_subnet.called_once_with()
        dns_mock.DnsController.create_dns_zones.called_once_with()
        self.assertIsInstance(ipam_subnet, drv.InfobloxSubnet)

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_update_subnet_no_zone_change(self, ib_cxt_mock, ipam_mock,
                                          dns_mock):
        pools = [netaddr.ip.IPRange('192.168.10.3', '192.168.10.25')]
        driver = self._mock_driver(requested_pools=pools)
        driver._grid_config.default_domain_name_pattern = (
            '{subnet_id}.cloud.global.com')
        driver._get_ib_network = mock.Mock(return_value=mock.Mock())
        subnet_factory = driver.get_subnet_request_factory()
        subnet_request = subnet_factory.get_request(self.ctx,
                                                    driver._fetch_subnet(),
                                                    None)

        driver.update_subnet(subnet_request)

        ipam_mock.IpamSyncController.update_subnet_allocation_pools.\
            called_once_with()
        assert not ipam_mock.IpamSyncController.create_dns_zones.called
        ipam_mock.IpamSyncController.get_subnet.called_once_with()
        ipam_mock.IpamSyncController.update_subnet_eas.called_once_with()

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_update_subnet_zone_change(self, ib_cxt_mock, ipam_mock, dns_mock):
        pools = [netaddr.ip.IPRange('192.168.10.3', '192.168.10.25')]
        driver = self._mock_driver(requested_pools=pools)
        subnet_factory = driver.get_subnet_request_factory()
        driver._grid_config.default_domain_name_pattern = (
            '{subnet_name}.cloud.global.com')
        driver._get_ib_network = mock.Mock(return_value=mock.Mock())
        test_subnet = driver._fetch_subnet()
        test_subnet['name'] = 'subnet-name-new'
        subnet_request = subnet_factory.get_request(self.ctx,
                                                    test_subnet,
                                                    None)

        driver.update_subnet(subnet_request)

        ipam_mock.IpamSyncController.update_subnet_allocation_pools.\
            called_once_with()
        ipam_mock.IpamSyncController.create_dns_zones.called_once_with()
        dns_mock.DnsController.create_dns_zones.called_once_with()
        ipam_mock.IpamSyncController.update_subnet_eas.called_once_with()

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_remove_subnet(self, ib_cxt_mock, ipam_mock, dns_mock):
        driver = self._mock_driver()
        subnet_id = 'subnet-id'
        driver._get_ib_network = mock.Mock(return_value=mock.Mock())

        driver.remove_subnet(subnet_id)

        ipam_mock.IpamSyncController.delete_subnet.called_once_with()
        dns_mock.DnsController.delete_dns_zones.called_once_with()

    def _mock_port(self, device_owner=None):
        return {'id': 'port-id',
                'name': 'port-name',
                'mac_address': 'mac-address',
                'tenant_id': 'tenant-id',
                'device_id': 'device-id',
                'device_owner': device_owner}

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_allocate_specific_ip(self, ib_cxt_mock, ipam_mock, dns_mock):
        driver = self._mock_driver()
        subnet_id = 'subnet-id'
        ipam_subnet = driver.get_subnet(subnet_id)

        port = self._mock_port()
        expected_ip = {'ip_address': '192.168.1.15'}
        address_factory = driver.get_address_request_factory()
        address_request = address_factory.get_request(self.ctx,
                                                      port,
                                                      expected_ip)

        ipam_subnet.allocate(address_request)

        ipam_mock.IpamSyncController.allocate_specific_ip.called_once_with(
            expected_ip['ip_address'],
            port['mac_address'],
            port['id'],
            port['tenant_id'],
            port['device_id'],
            port['device_owner'])
        dns_mock.DnsController.bind_names.called_once_with(
            expected_ip['ip_address'],
            None,
            port['id'],
            port['tenant_id'],
            port['device_id'],
            port['device_owner'])

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    def test_allocate_ip_from_pool(self, ib_cxt_mock, ipam_mock, dns_mock):
        driver = self._mock_driver()
        subnet_id = 'subnet-id'
        ipam_subnet = driver.get_subnet(subnet_id)

        port = self._mock_port()
        address_factory = driver.get_address_request_factory()
        address_request = address_factory.get_request(self.ctx,
                                                      port,
                                                      {})

        allocated_ip = ipam_subnet.allocate(address_request)

        ipam_mock.IpamSyncController.allocate_ip_from_pool.called_once_with(
            ipam_subnet._neutron_subnet['id'],
            ipam_subnet._neutron_subnet['allocation_pools'],
            port['mac_address'],
            port['id'],
            port['tenant_id'],
            port['device_id'],
            port['device_owner'])
        dns_mock.DnsController.bind_names.called_once_with(
            allocated_ip,
            None,
            port['id'],
            port['tenant_id'],
            port['device_id'],
            port['device_owner'])

    @mock.patch('networking_infoblox.neutron.common.dns.DnsController')
    @mock.patch('networking_infoblox.neutron.common.ipam.IpamSyncController')
    @mock.patch('networking_infoblox.neutron.common.context.InfobloxContext')
    @mock.patch('infoblox_client.objects.FixedAddress')
    def test_deallocate_ip(self, fixed_ip_mock, ib_cxt_mock, ipam_mock,
                           dns_mock):
        driver = self._mock_driver()
        subnet_id = 'subnet-id'
        ipam_subnet = driver.get_subnet(subnet_id)

        expected_ip = {'ip_address': '192.168.1.15'}

        port = self._mock_port()
        address_factory = driver.get_address_request_factory()
        address_request = address_factory.get_request(self.ctx,
                                                      port,
                                                      expected_ip)

        fixed_address_mock = mock.Mock()
        fixed_ip_mock.search = mock.Mock(return_value=fixed_address_mock)
        ipam_subnet._build_address_request_from_ib_address = mock.Mock(
            return_value=address_request)

        ipam_subnet.deallocate(netaddr.IPAddress(expected_ip['ip_address']))

        ipam_mock.IpamSyncController.deallocate_ip.called_once_with(
            expected_ip['ip_address'])
        dns_mock.DnsController.unbind_names.called_once_with(
            expected_ip['ip_address'],
            None,
            port['id'],
            port['tenant_id'],
            port['device_id'],
            port['device_owner'])


class FakePool(object):

    def __init__(self, object_mock):
        self.mock = object_mock

    @drv.rollback_wrapper
    def allocate_something(self, rollback_list, fail=False):
        rollback_list.append(self.mock)
        if fail:
            raise ValueError


class TestWrapper(base.TestCase):

    def test_rollback_wrapper(self):
        created_object = mock.Mock()
        pool = FakePool(created_object)
        pool.allocate_something()
        self.assertEqual(False, created_object.delete.called)

    def test_rollback_wrapper_on_failure(self):
        created_object = mock.Mock()
        pool = FakePool(created_object)
        self.assertRaises(ValueError, pool.allocate_something, fail=True)
        self.assertEqual(True, created_object.delete.called)

    def test_rollback_wrapper_on_delete_failure(self):
        created_object = mock.Mock()
        created_object.delete.side_effect = ib_exc.InfobloxException(
            'error_response')
        pool = FakePool(created_object)
        self.assertRaises(ValueError, pool.allocate_something, fail=True)
        self.assertEqual(True, created_object.delete.called)



# Copyright 2016 Infoblox Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""network_name_change_support

Revision ID: 0075c5a73439
Revises: 5267e17f370f
Create Date: 2016-08-09 16:19:00.207834

"""

# revision identifiers, used by Alembic.
revision = '0075c5a73439'
down_revision = '5267e17f370f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'infoblox_networks',
        sa.Column('network_id', sa.String(64),
                  nullable=False, primary_key=True),
        sa.Column('network_name', sa.String(255), nullable=False),
    )



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions

from networking_infoblox._i18n import _


class InfobloxNeutronException(exceptions.NeutronException):
    """Generic Infoblox Exception."""
    def __init__(self, response, **kwargs):
        self.response = response
        super(InfobloxNeutronException, self).__init__(**kwargs)


class InfobloxInvalidCloudDataCenter(exceptions.NeutronException):
    message = _("Invalid cloud data center: '%(data_center_id)s'")


class InfobloxCannotFindMember(exceptions.NeutronException):
    message = _("Cannot find the member: '%(member)s'")


class InfobloxCannotReserveAuthorityMember(exceptions.NeutronException):
    message = _("Cannot reserve the authority member for network view: "
                "%(network_view)s")


class InfobloxAuthorityMemberNotReserved(exceptions.NeutronException):
    message = _("Authority member is not reserved for network view: "
                "%(network_view)s")


class InfobloxDHCPMemberNotReserved(exceptions.NeutronException):
    message = _("DHCP member is not reserved for network view: "
                "(%(network_view)s), cidr (%(cidr)s)")


class InfobloxDNSMemberNotReserved(exceptions.NeutronException):
    message = _("DNS member is not reserved for network view "
                "(%(network_view)s), cidr (%(cidr)s)")


class InfobloxNetworkViewMappingNotFound(exceptions.NeutronException):
    message = _("Cannot find a network view mapped for subnet %(subnet_id)s")


class MultipleNetworkViewMappingFound(exceptions.Conflict):
    message = _("Multiple network view mapping found. You need to add more "
                "filters to narrow down the search")


class InfobloxNetworkViewNotFound(exceptions.NeutronException):
    message = _("Network view '%(network_view)s' does not exist.")


class InfobloxNetworkViewNotParticipated(exceptions.NeutronException):
    message = _("Network view '%(network_view)s' found but not participated.")


class InfobloxDefaultNetworkViewNotFound(exceptions.NeutronException):
    message = _("No default network view exists.")


class InfobloxCannotCreateSubnet(exceptions.NeutronException):
    message = _("Cannot create a subnet because %(reason)s")


class InfobloxCannotFindSubnet(exceptions.NeutronException):
    message = _("Cannot find the subnet %(subnet_id)s for %(cidr)s from NIOS")


class InfobloxPrivateSubnetAlreadyExist(exceptions.Conflict):
    message = _("Network with the same CIDR already exists on NIOS.")


class InfobloxCannotFindFixedIp(exceptions.NeutronException):
    message = _("Cannot find the fixed IP %(ip)s")


class InfobloxClientException(exceptions.NeutronException):
    message = _("InfobloxClientException '%(msg)s'")


class InfobloxValueError(exceptions.NeutronException):
    message = _("InfobloxValueError '%(msg)s' "
                "Refer to neutron log for more detail")



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_const

from neutron.ipam import requests


class InfobloxSubnetRequestFactory(requests.SubnetRequestFactory):
    """Infoblox Address Request Factory

    Introduce custom address request types specific for Infoblox IPAM Driver
    """
    @classmethod
    def get_request(cls, context, subnet, subnetpool):
        request = super(InfobloxSubnetRequestFactory, cls).get_request(
            context, subnet, subnetpool)
        request.name = subnet.get('name')
        # update_subnet does not pass network_id. community code can be
        # improved to include it
        request.network_id = subnet.get('network_id')
        request.subnetpool_id = subnetpool['id'] if subnetpool else None
        request.enable_dhcp = subnet.get('enable_dhcp')
        # neutron api sets optional attributes to the following
        #     ATTR_NOT_SPECIFIED = object()
        dns_nameservers = subnet.get('dns_nameservers')
        if dns_nameservers is n_const.ATTR_NOT_SPECIFIED:
            dns_nameservers = []
        request.dns_nameservers = dns_nameservers
        return request


class InfobloxAddressRequestFactory(requests.AddressRequestFactory):
    """Infoblox Address Request Factory

    Introduce custom address request types specific for Infoblox IPAM Driver
    """
    @classmethod
    def get_request(cls, context, port, ip_dict):

        request = super(InfobloxAddressRequestFactory, cls).get_request(
            context, port, ip_dict)
        request.mac = port.get('mac_address')
        request.tenant_id = port.get('tenant_id') or context.tenant_id
        request.port_id = port.get('id')
        request.port_name = port.get('name')
        request.device_id = port.get('device_id')
        request.device_owner = port.get('device_owner')
        return request


class InfobloxFixedAddressRequest(requests.SpecificAddressRequest):

    def __init__(self, address, tenant_id, mac, port_id,
                 device_id, device_owner):
        super(InfobloxFixedAddressRequest, self).__init__(address)
        self.tenant_id = tenant_id
        self.mac = mac
        self.port_id = port_id
        self.device_id = device_id
        self.device_owner = device_owner


class InfobloxFloatingAddressRequest(InfobloxFixedAddressRequest):
    """Used to request allocating the floating ip address."""


class InfobloxRouterGatewayAddressRequest(InfobloxFixedAddressRequest):
    """Used to request allocating the special router gateway address."""


class InfobloxDhcpPortAddressRequest(InfobloxFixedAddressRequest):
    """Used to request allocating the dhcp port address."""


class InfobloxAnyAddressRequest(requests.AnyAddressRequest):
    """Used to request allocating any address from subnet."""

    def __init__(self, tenant_id, mac, port_id, device_id, device_owner):
        super(InfobloxAnyAddressRequest, self).__init__()
        self.tenant_id = tenant_id
        self.mac = mac
        self.port_id = port_id
        self.device_id = device_id
        self.device_owner = device_owner


class InfobloxAutomaticAddressRequest(requests.AutomaticAddressRequest):
    """Used to request automatic address for auto-addressed subnets."""

    def __init__(self, tenant_id, mac, port_id,
                 device_id, device_owner, prefix):
        super(InfobloxAutomaticAddressRequest, self).__init__(prefix=prefix,
                                                              mac=mac)
        self.tenant_id = tenant_id
        self.mac = mac
        self.port_id = port_id
        self.device_id = device_id
        self.device_owner = device_owner


class InfobloxAddressRequestFactoryV2(requests.AddressRequestFactory):
    """Infoblox Address Request Factory.

    Introduce custom address request types specific for Infoblox IPAM Driver
    """

    @classmethod
    def get_request(cls, context, port, ip_dict):
        tenant_id = port.get('tenant_id') or context.tenant_id
        if ip_dict.get('ip_address'):
            if port['device_owner'] == n_const.DEVICE_OWNER_DHCP:
                request_class = InfobloxDhcpPortAddressRequest
            elif port['device_owner'] == n_const.DEVICE_OWNER_ROUTER_GW:
                request_class = InfobloxRouterGatewayAddressRequest
            elif port['device_owner'] == n_const.DEVICE_OWNER_FLOATINGIP:
                request_class = InfobloxFloatingAddressRequest
            else:
                request_class = InfobloxFixedAddressRequest
            return request_class(ip_dict['ip_address'],
                                 tenant_id,
                                 port['mac_address'],
                                 port['id'],
                                 port['device_id'],
                                 port['device_owner'])
        elif ip_dict.get('eui64_address'):
            return InfobloxAutomaticAddressRequest(
                tenant_id,
                port['mac_address'],
                port['id'],
                port['device_id'],
                port['device_owner'],
                ip_dict['subnet_cidr'])
        else:
            return InfobloxAnyAddressRequest(tenant_id,
                                             port['mac_address'],
                                             port['id'],
                                             port['device_id'],
                                             port['device_owner'])



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from datetime import timedelta
from oslo_log import log as logging
import six
import socket

from neutron_lib import context as neutron_context

from infoblox_client import connector
from infoblox_client import objects as ib_objects

from networking_infoblox._i18n import _LI
from networking_infoblox.neutron.common import config as cfg
from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import exceptions as exc
from networking_infoblox.neutron.common import mapping as grid_mapping
from networking_infoblox.neutron.common import member as grid_member
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi


LOG = logging.getLogger(__name__)


class GridSyncer(object):
    def __init__(self):
        self._context = neutron_context.get_admin_context()
        self._grid_manager = GridManager(self._context)

    def is_sync_needed(self, interval):
        return self._grid_manager.is_sync_needed(interval)

    def sync(self, force_sync=False):
        self._grid_manager.sync(force_sync)


class GridManager(object):

    grid_config = None
    member = None
    mapping = None
    last_sync_time = None

    def __init__(self, context):
        self.grid_config = self._create_grid_configuration(context)
        self.member = grid_member.GridMemberManager(self.grid_config)
        self.mapping = grid_mapping.GridMappingManager(self.grid_config)
        self.hostname = socket.gethostname()

    def is_sync_needed(self, resync_interval):
        session = self.grid_config.context.session
        self.last_sync_time = dbi.get_last_sync_time(session)
        return (not self.last_sync_time or
                (datetime.utcnow() - self.last_sync_time > timedelta(
                    seconds=resync_interval)))

    def sync(self, force_sync=False):
        """Synchronize members, config, and mapping between NIOS and neutron.

        First sync members, then config, and lastly mapping because config
        sync needs GM info and mapping needs grid config.
        """
        session = self.grid_config.context.session
        allow_sync = False
        if self.grid_config.grid_sync_support:
            interval = self.grid_config.grid_sync_minimum_wait_time
            if force_sync or self.is_sync_needed(interval):
                allow_sync = True

        if allow_sync:
            self.member.sync()
            self.grid_config.sync()
            self.mapping.sync()
            self.last_sync_time = datetime.utcnow().replace(microsecond=0)
            dbi.record_last_sync_time(session, self.last_sync_time)
            self._report_sync_time()
            LOG.info("Infoblox grid has been synced up.")

    def get_config(self):
        """Gets grid configuration.

        Before this call is made, the grid member must be in sync.
        """
        self.grid_config.sync()

    @staticmethod
    def _create_grid_configuration(context):
        grid_conf = GridConfiguration(context)
        grid_conf.grid_id = cfg.CONF.infoblox.cloud_data_center_id
        grid_opts = cfg.get_infoblox_grid_opts(grid_conf.grid_id)
        if not grid_opts['grid_master_host']:
            raise exc.InfobloxInvalidCloudDataCenter(
                data_center_id=grid_conf.grid_id)
        grid_conf.grid_name = grid_opts['data_center_name']
        grid_conf.grid_master_host = grid_opts['grid_master_host']
        grid_conf.grid_master_name = grid_opts['grid_master_name']
        grid_conf.admin_user_name = grid_opts['admin_user_name']
        grid_conf.admin_password = grid_opts['admin_password']
        grid_conf.wapi_version = grid_opts['wapi_version']
        grid_conf.ssl_verify = grid_opts['ssl_verify']
        grid_conf.http_request_timeout = grid_opts['http_request_timeout']
        grid_conf.http_pool_connections = grid_opts['http_pool_connections']
        grid_conf.http_pool_maxsize = grid_opts['http_pool_maxsize']
        grid_conf.wapi_max_results = grid_opts['wapi_max_results']

        # cloud user is used as admin, it needs to have proper permissions to
        # deal with non-delegated objects.
        gm_connection_opts = {
            'host': grid_conf.grid_master_host,
            'username': grid_conf.admin_user_name,
            'password': grid_conf.admin_password,
            'wapi_version': grid_conf.wapi_version,
            'ssl_verify': grid_conf.ssl_verify,
            'log_api_calls_as_info': True,
            'http_request_timeout': grid_conf.http_request_timeout,
            'http_pool_connections': grid_conf.http_pool_connections,
            'http_pool_maxsize': grid_conf.http_pool_maxsize,
            'max_results': grid_conf.wapi_max_results}
        # Silent ssl warnings, if certificate verification is not enabled
        if gm_connection_opts['ssl_verify'] == 'False':
            gm_connection_opts['silent_ssl_warnings'] = True
        grid_conf.gm_connector = connector.Connector(gm_connection_opts)
        return grid_conf

    def _report_sync_time(self):
        if self.grid_config.report_grid_sync_time is False:
            return

        conn = self.grid_config.gm_connector
        gm = self.grid_config.get_gm_member()
        if gm.member_ip:
            gm = ib_objects.Member.search(conn, ipv4_address=gm.member_ip)
        else:
            gm = ib_objects.Member.search(conn, ipv6_address=gm.member_ipv6)

        sync_info = (str(self.grid_config.grid_id) + ':' +
                     self.hostname + ' => ' +
                     self.last_sync_time.strftime("%Y-%m-%d %H:%M:%S"))

        sync_info_list = gm.extattrs.get(const.EA_LAST_GRID_SYNC_TIME)
        if sync_info_list:
            # if a single entry exits. NIOS returns as string rather than list.
            if isinstance(sync_info_list, six.string_types):
                sync_info_list = [sync_info_list]

            found_sync_idx_lst = [idx for idx, si in enumerate(sync_info_list)
                                  if self.hostname in si]
            if found_sync_idx_lst:
                sync_info_list[found_sync_idx_lst[0]] = sync_info
            else:
                sync_info_list.append(sync_info)
        else:
            sync_info_list = [sync_info]

        gm.extattrs.set(const.EA_LAST_GRID_SYNC_TIME, sync_info_list)
        gm.update()


class GridConfiguration(object):

    property_to_ea_mapping = {
        'grid_sync_support':
            const.EA_GRID_CONFIG_GRID_SYNC_SUPPORT,
        'grid_sync_minimum_wait_time':
            const.EA_GRID_CONFIG_GRID_SYNC_MINIMUM_WAIT_TIME,
        'grid_sync_maximum_wait_time':
            const.EA_GRID_CONFIG_GRID_SYNC_MAXIMUM_WAIT_TIME,
        'default_network_view_scope':
            const.EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW_SCOPE,
        'default_network_view': const.EA_GRID_CONFIG_DEFAULT_NETWORK_VIEW,
        'default_host_name_pattern':
            const.EA_GRID_CONFIG_DEFAULT_HOST_NAME_PATTERN,
        'default_domain_name_pattern':
            const.EA_GRID_CONFIG_DEFAULT_DOMAIN_NAME_PATTERN,
        'external_host_name_pattern':
            const.EA_GRID_CONFIG_EXTERNAL_HOST_NAME_PATTERN,
        'external_domain_name_pattern':
            const.EA_GRID_CONFIG_EXTERNAL_DOMAIN_NAME_PATTERN,
        'ns_group': const.EA_GRID_CONFIG_NS_GROUP,
        'dns_view': const.EA_GRID_CONFIG_DNS_VIEW,
        'network_template': const.EA_GRID_CONFIG_NETWORK_TEMPLATE,
        'admin_network_deletion': const.EA_GRID_CONFIG_ADMIN_NETWORK_DELETION,
        'ip_allocation_strategy': const.EA_GRID_CONFIG_IP_ALLOCATION_STRATEGY,
        'dns_record_binding_types':
            const.EA_GRID_CONFIG_DNS_RECORD_BINDING_TYPES,
        'dns_record_unbinding_types':
            const.EA_GRID_CONFIG_DNS_RECORD_UNBINDING_TYPES,
        'dns_record_removable_types':
            const.EA_GRID_CONFIG_DNS_RECORD_REMOVABLE_TYPES,
        'dhcp_support': const.EA_GRID_CONFIG_DHCP_SUPPORT,
        'dns_support': const.EA_GRID_CONFIG_DNS_SUPPORT,
        'relay_support': const.EA_GRID_CONFIG_RELAY_SUPPORT,
        'use_grid_master_for_dhcp': const.EA_GRID_CONFIG_USE_GM_FOR_DHCP,
        'report_grid_sync_time': const.EA_GRID_CONFIG_REPORT_GRID_SYNC_TIME,
        'allow_service_restart': const.EA_GRID_CONFIG_ALLOW_SERVICE_RESTART,
        'allow_static_zone_deletion':
            const.EA_GRID_CONFIG_ALLOW_STATIC_ZONE_DELETION,
        'zone_creation_strategy': const.EA_GRID_CONFIG_ZONE_CREATION_STRATEGY,
        'tenant_name_persistence': const.EA_GRID_CONFIG_TENANT_NAME_PERSISTENCE
    }

    def __init__(self, context):
        self.context = context

        # grid info from neutron conf
        self.grid_id = None
        self.grid_name = None
        self.grid_master_host = None
        self.grid_master_name = None

        # grid connection from neutron conf
        self.admin_user_name = None
        self.admin_password = None
        self.ssl_verify = False
        self.http_request_timeout = 120
        self.http_pool_connections = 100
        self.http_pool_maxsize = 100

        # connector object to GM
        self.gm_connector = None
        self._wapi_version = None
        self._is_cloud_wapi = False

        # default settings from nios grid master
        self._set_default_values()

    @property
    def wapi_version(self):
        return self._wapi_version

    @wapi_version.setter
    def wapi_version(self, value):
        self._wapi_version = value
        if value:
            self._is_cloud_wapi =\
                utils.get_features(self.wapi_version).cloud_api

    @property
    def is_cloud_wapi(self):
        return self._is_cloud_wapi

    def sync(self):
        discovered_config = self._discover_config(self.get_gm_member())
        if discovered_config:
            self._update_fields(discovered_config)
            LOG.debug(_LI("grid config synced: %s"), self.__dict__)

    def get_grid_connection(self):
        grid_connection = {
            "wapi_version": self.wapi_version,
            "ssl_verify": self.ssl_verify,
            "http_pool_connections": self.http_pool_connections,
            "http_pool_maxsize": self.http_pool_maxsize,
            "http_request_timeout": self.http_request_timeout,
            "admin_user": {"name": self.admin_user_name,
                           "password": self.admin_password}
        }
        return grid_connection

    def _discover_config(self, gm_member):
        return_fields = ['extattrs']
        if utils.get_features(self.wapi_version).member_ipv6_setting:
            return_fields.append('ipv6_setting')

        obj_type = 'member'
        payload = {'host_name': gm_member['member_name']}
        config = self.gm_connector.get_object(
            obj_type, payload=payload, return_fields=return_fields)
        return config[0] if config and config[0].get('extattrs') else None

    def _set_default_values(self):
        for prop, key in self.property_to_ea_mapping.items():
            if key in const.GRID_CONFIG_DEFAULTS:
                setattr(self, prop, const.GRID_CONFIG_DEFAULTS[key])

    def _update_fields(self, extattr):
        for pm in self.property_to_ea_mapping:
            self._update_from_ea(pm,
                                 self.property_to_ea_mapping[pm],
                                 extattr)

    def _update_from_ea(self, field, ea_name, extattrs):
        value = utils.get_ea_value(ea_name, extattrs,
                                   ea_name in const.EA_MULTI_VALUES)
        if value:
            setattr(self, field, self._value_to_bool(value))

    @staticmethod
    def _value_to_bool(value):
        """Converts value returned by NIOS into boolean if possible."""
        if value == 'True':
            return True
        elif value == 'False':
            return False
        return value

    def get_gm_member(self):
        session = self.context.session
        members = dbi.get_members(session,
                                  grid_id=self.grid_id,
                                  member_type=const.MEMBER_TYPE_GRID_MASTER,
                                  member_status=const.MEMBER_STATUS_ON)
        if not members or len(members) != 1:
            raise exc.InfobloxCannotFindMember(member="GM")
        return members[0]



# Copyright (c) 2016 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from networking_infoblox.neutron.common import keystone_manager

from networking_infoblox.tests import base


class DummyInfobloxOpts(object):
    def __init__(self):
        self.keystone_auth_uri = 'auth_uri'
        self.keystone_admin_username = 'cloud_admin_user'
        self.keystone_admin_password = 'cloud_admin_password'
        self.keystone_admin_project_name = 'cloud_admin_project'
        self.keystone_admin_tenant_name = 'cloud_admin_tenant'
        self.keystone_admin_user_domain_id = 'cloud_admin_domain'
        self.keystone_admin_project_domain_id = 'cloud_admin_domain'
        self.keystone_auth_version = 'v2.0'
        self.keystone_admin_domain_id = ''


class TestKeystoneManager(base.TestCase):

    def setUp(self):
        super(TestKeystoneManager, self).setUp()
        self.tenants = []
        for i in range(1, 4):
            tenant = mock.Mock()
            tenant.id = str(i)
            tenant.tenant_id = tenant.id
            tenant.name = 'tenant_%s' % i
            self.tenants.append(tenant)
        self.networks = [
            {'tenant_id': '1'},
            {'tenant_id': '2'},
            {'tenant_id': '3'},
            {'tenant_id': '4'},
            ]

    def test_get_identity_service_with_auth_version(self):
        version = 'v3'
        ib_opts = DummyInfobloxOpts()
        ib_opts.keystone_auth_version = version
        identity = ib_opts.keystone_auth_uri + '/' + (
            ib_opts.keystone_auth_version)
        # call tested function
        r_identity, r_version = keystone_manager.get_identity_service(ib_opts)
        assert r_identity == identity
        assert r_version == version

    def test_get_identity_service_without_auth_version(self):
        ib_opts = DummyInfobloxOpts()
        version = 'v3'
        auth_url = 'auth_url' + '/' + version
        ib_opts.keystone_auth_uri = auth_url
        # call tested function
        r_identity, r_version = keystone_manager.get_identity_service(ib_opts)
        assert r_identity == auth_url
        assert r_version == version

    @mock.patch('keystoneauth1.loading.load_session_from_conf_options')
    @mock.patch('keystoneauth1.identity.generic.Password')
    def _test_get_keystone_client(self, ClientMock, ConfMock,
                                  PasswordMock, loadSessionMock):
        ib_opts = ConfMock.infoblox
        session = 'session'
        auth = 'auth'
        auth_url = ib_opts.keystone_auth_uri + '/' + (
            ib_opts.keystone_auth_version)
        loadSessionMock.return_value = session
        PasswordMock.return_value = auth
        # call tested function
        k_client = keystone_manager.get_keystone_client()
        PasswordMock.assert_called_once_with(
            auth_url=auth_url,
            username=ib_opts.keystone_admin_username,
            password=ib_opts.keystone_admin_password,
            project_name=ib_opts.keystone_admin_project_name,
            user_domain_id=ib_opts.keystone_admin_user_domain_id,
            project_domain_id=ib_opts.keystone_admin_project_domain_id,
            tenant_name=ib_opts.keystone_admin_tenant_name,
            domain_id=ib_opts.keystone_admin_domain_id)
        loadSessionMock.assert_called_once_with(ConfMock, 'infoblox',
                                                auth=auth)
        ClientMock.assert_called_once_with(session=session)
        assert k_client == ClientMock.return_value

    @mock.patch('networking_infoblox.neutron.common.keystone_manager.CONF')
    @mock.patch('keystoneclient.v2_0.client.Client')
    def test_get_keystone_client_v2_0(self, ClientMock, ConfMock):
        ClientMock.return_value = 'keystone_client_v2_0'
        # prepare mocks
        ConfMock.infoblox = DummyInfobloxOpts()
        self._test_get_keystone_client(ClientMock, ConfMock)

    @mock.patch('networking_infoblox.neutron.common.keystone_manager.CONF')
    @mock.patch('keystoneclient.v3.client.Client')
    def test_get_keystone_client_v3(self, ClientMock, ConfMock):
        ClientMock.return_value = 'keystone_client_v3'
        # prepare mocks
        ConfMock.infoblox = DummyInfobloxOpts()
        ConfMock.infoblox.keystone_auth_version = 'v3'
        self._test_get_keystone_client(ClientMock, ConfMock)

    @mock.patch('networking_infoblox.neutron.common.keystone_manager.'
                'get_keystone_client')
    def _test_get_all_tenants(self, version, get_keystone_client_mock):
        # prepare mocks
        client_mock = mock.MagicMock()
        client_mock.version = version
        get_keystone_client_mock.return_value = client_mock
        # call tested function
        keystone_manager.get_all_tenants()
        # check calls
        get_keystone_client_mock.assert_called_once_with()
        if version == 'v3':
            client_mock.projects.list.assert_called_once_with()
        else:
            client_mock.tenants.list.assert_called_once_with()

    def test_get_all_tenants_v2_0(self):
        self._test_get_all_tenants('v2.0')

    def test_get_all_tenants_v3(self):
        self._test_get_all_tenants('v3')

    @mock.patch('networking_infoblox.neutron.common.keystone_manager.'
                'sync_tenants_from_keystone')
    @mock.patch('networking_infoblox.neutron.db.infoblox_db.'
                'add_or_update_tenant')
    @mock.patch('networking_infoblox.neutron.db.infoblox_db.get_tenants')
    def _test_update_tenant_mapping(self, networks, tenants,
                                    expected_results,
                                    get_tenants, add_or_update_tenant,
                                    sync_tenants_from_keystone):
        context = mock.Mock()
        context.session = 'test_session'
        tenant_id = '1'
        tenant_name = 'test_tenant_name_1'
        get_unknown = mock.Mock()
        get_unknown_params = []
        func = keystone_manager._get_unknown_ids_from_dict

        def store_params(param):
            get_unknown_params.append(param.copy())
            return func(param)

        get_unknown.side_effect = store_params
        get_tenants.return_value = tenants
        with mock.patch('networking_infoblox.neutron.common.keystone_manager.'
                        '_get_unknown_ids_from_dict', get_unknown):
            keystone_manager.update_tenant_mapping(
                context, networks, tenant_id, tenant_name)
        add_or_update_tenant.assert_called_once_with(context.session,
                                                     tenant_id, tenant_name)
        assert get_unknown.call_count == expected_results['get_unknown_count']
        assert get_unknown_params == expected_results['get_unknown_params']
        if expected_results['get_tenants_called']:
            get_tenants.assert_called_once()
            assert get_tenants.call_args[0] == (context.session,)
            tenant_ids = expected_results['get_tenant_tenant_ids']
            assert(
                sorted(
                    get_tenants.call_args[1]['tenant_ids']) == tenant_ids)
        else:
            get_tenants.assert_not_called()
        if expected_results['sync_tenants_from_keystone_called']:
            sync_tenants_from_keystone.assert_called_once_with(context)
        else:
            sync_tenants_from_keystone.assert_not_called()

    def test_update_tenant_mapping(self):
        networks = self.networks
        tenants = self.tenants
        expected_results = {
            'get_unknown_count': 2,
            'get_unknown_params': [
                {
                    networks[0]['tenant_id']: False,
                    networks[1]['tenant_id']: True,
                    networks[2]['tenant_id']: True,
                    networks[3]['tenant_id']: True
                },
                {
                    networks[0]['tenant_id']: False,
                    networks[1]['tenant_id']: False,
                    networks[2]['tenant_id']: False,
                    networks[3]['tenant_id']: True
                }],
            'get_tenants_called': True,
            'get_tenant_tenant_ids': ['2', '3', '4'],
            'sync_tenants_from_keystone_called': True
            }
        self._test_update_tenant_mapping(networks, tenants,
                                         expected_results)

    def test_update_tenant_mapping_all_tenant_in_db(self):
        networks = self.networks
        networks[3]['tenant_id'] = networks[0]['tenant_id']
        tenants = self.tenants
        expected_results = {
            'get_unknown_count': 2,
            'get_unknown_params': [
                {
                    networks[0]['tenant_id']: False,
                    networks[1]['tenant_id']: True,
                    networks[2]['tenant_id']: True,
                    networks[3]['tenant_id']: False
                },
                {
                    networks[0]['tenant_id']: False,
                    networks[1]['tenant_id']: False,
                    networks[2]['tenant_id']: False,
                    networks[3]['tenant_id']: False
                }],
            'get_tenants_called': True,
            'get_tenant_tenant_ids': ['2', '3'],
            'sync_tenants_from_keystone_called': False
            }
        self._test_update_tenant_mapping(networks, tenants,
                                         expected_results)

    def test_update_tenant_mapping_all_tenant_in_networks(self):
        networks = self.networks
        for i in range(1, len(networks)):
            networks[i]['tenant_id'] = networks[0]['tenant_id']
        tenants = self.tenants
        expected_results = {
            'get_unknown_count': 1,
            'get_unknown_params': [
                {
                    networks[0]['tenant_id']: False,
                    networks[1]['tenant_id']: False,
                    networks[2]['tenant_id']: False,
                    networks[3]['tenant_id']: False
                }],
            'get_tenants_called': False,
            'sync_tenants_from_keystone_called': False
            }
        self._test_update_tenant_mapping(networks, tenants,
                                         expected_results)

    @mock.patch('networking_infoblox.neutron.common.keystone_manager.'
                'get_all_tenants')
    @mock.patch('networking_infoblox.neutron.db.infoblox_db.'
                'add_or_update_tenant')
    def test_sync_tenants_from_keystone(self, AddTenantMock, get_all_tenants):
        # prepare test data
        context = mock.Mock()
        context.session = 'test_session'
        get_all_tenants.return_value = self.tenants
        # call tested function
        ret = keystone_manager.sync_tenants_from_keystone(context)
        # check return value and calls
        assert ret == len(self.tenants)
        get_all_tenants.assert_called_once_with()
        expected_call_list = [
            mock.call(context.session, t.id, t.name) for t in self.tenants]
        assert AddTenantMock.call_args_list == expected_call_list



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import netaddr
from neutron_lib import constants as n_const

from networking_infoblox.ipam import requests
from networking_infoblox.tests import base


class TestDriver(base.TestCase):

    def _get_mocked_port(self, device_owner=None, mac=None):
        if device_owner is None:
            device_owner = mock.Mock()
        if mac is None:
            mac = mock.Mock()
        return {'mac_address': mac,
                'tenant_id': mock.Mock(),
                'id': mock.Mock(),
                'device_id': mock.Mock(),
                'device_owner': device_owner}

    def _validate_ib_fields_in_request(self, port, request):
        self.assertEqual(port['mac_address'], request.mac)
        self.assertEqual(port['id'], request.port_id)
        self.assertEqual(port['device_id'], request.device_id)
        self.assertEqual(port['device_owner'], request.device_owner)

    def _test_static_requests(self, device_owner, request_class):
        context = mock.Mock()
        port = self._get_mocked_port(device_owner=device_owner)
        ip_dict = {'ip_address': '192.168.1.10',
                   'subnet_id': mock.Mock()}
        req = requests.InfobloxAddressRequestFactoryV2.get_request(context,
                                                                   port,
                                                                   ip_dict)
        self.assertIsInstance(req, request_class)
        self.assertEqual(req.address,
                         netaddr.IPAddress(ip_dict['ip_address']))
        self._validate_ib_fields_in_request(port, req)

    def test_fixed_address_request(self):
        self._test_static_requests(mock.Mock(),
                                   requests.InfobloxFixedAddressRequest)

    def test_floating_address_request(self):
        self._test_static_requests(n_const.DEVICE_OWNER_FLOATINGIP,
                                   requests.InfobloxFloatingAddressRequest)

    def test_router_gateway_address_request(self):
        self._test_static_requests(
            n_const.DEVICE_OWNER_ROUTER_GW,
            requests.InfobloxRouterGatewayAddressRequest)

    def test_dhcp_port_address_request(self):
        self._test_static_requests(
            n_const.DEVICE_OWNER_DHCP,
            requests.InfobloxDhcpPortAddressRequest)

    def test_any_address_request(self):
        context = mock.Mock()
        port = self._get_mocked_port()
        ip_dict = {}
        req = requests.InfobloxAddressRequestFactoryV2.get_request(context,
                                                                   port,
                                                                   ip_dict)
        self.assertIsInstance(req, requests.InfobloxAnyAddressRequest)
        self._validate_ib_fields_in_request(port, req)

    def test_auto_address_request(self):
        context = mock.Mock()
        mac = 'aa:ab:dc:11:22:33'
        port = self._get_mocked_port(mac=mac)
        ip_dict = {'eui64_address': True,
                   'subnet_cidr': 'fffe::/64',
                   'mac': mac}
        req = requests.InfobloxAddressRequestFactoryV2.get_request(context,
                                                                   port,
                                                                   ip_dict)
        self.assertIsInstance(req, requests.InfobloxAutomaticAddressRequest)
        self.assertEqual(req.address,
                         netaddr.IPAddress('fffe::a8ab:dcff:fe11:2233'))
        self._validate_ib_fields_in_request(port, req)



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import oslo_messaging
from oslo_service import loopingcall
from oslo_service import service

from neutron.agent import rpc as agent_rpc
from neutron.common import topics
from neutron_lib import context

from networking_infoblox._i18n import _LE
from networking_infoblox._i18n import _LI
from networking_infoblox._i18n import _LW
from networking_infoblox.neutron.common import config
from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import notification_handler


LOG = logging.getLogger(__name__)


class NotificationEndpoint(object):

    event_subscription_list = [
        # pre-committed ipam events
        'network.create.start',
        'subnet.create.start',
        # post-committed ipam events
        'network.create.end',
        'network.update.end',
        'network.delete.end',
        'subnet.create.end',
        'subnet.update.end',
        'subnet.delete.end',
        'port.create.end',
        'port.update.end',
        'port.delete.end',
        'floatingip.create.end',
        'floatingip.update.end',
        'floatingip.delete.end',
        # nova instance event
        'compute.instance.create.end',
        'compute.instance.delete.end']

    def __init__(self, context, grid_manager):
        self.context = context
        # Using filter in oslo_messaing 4.1.1 did not work for some reason
        # so commenting filter out
        # self.filter_rule = oslo_messaging.NotificationFilter(
        #    publisher_id='^(network|compute).*',
        #    event_type='|'.join(self.event_subscription_list))
        self.handler = notification_handler.IpamEventHandler(
            self.context, grid_manager=grid_manager)

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        if event_type in self.event_subscription_list:
            return self.handler.process(ctxt, publisher_id, event_type,
                                        payload, metadata)


class NotificationService(service.Service):
    """Listener for notification service."""

    NOTIFICATION_TOPIC = 'notifications'
    RESYNC_TRY_INTERVAL = 30

    def __init__(self, report_interval=None):
        super(NotificationService, self).__init__()
        self.report_thread = None
        self.event_listener = None
        self.executor = "blocking"
        if report_interval:
            self.report_interval = report_interval
        else:
            self.report_interval = config.CONF.AGENT.report_interval
        self.context = context.get_admin_context()
        # Make sure config is in sync before using grid_sync_maximum_wait_time
        self.grid_syncer = grid.GridSyncer()
        self.grid_syncer.sync(True)
        self.grid_manager = self.grid_syncer._grid_manager
        self._init_agent_report_thread()
        self._init_notification_listener()
        self._init_periodic_resync()

    def _init_notification_listener(self):
        self.transport = oslo_messaging.get_transport(config.CONF)
        self.event_targets = [
            oslo_messaging.Target(exchange=const.NOTIFICATION_EXCHANGE_NEUTRON,
                                  topic=self.NOTIFICATION_TOPIC),
            oslo_messaging.Target(exchange=const.NOTIFICATION_EXCHANGE_NOVA,
                                  topic=self.NOTIFICATION_TOPIC)
        ]
        self.event_endpoints = [NotificationEndpoint(self.context,
                                                     self.grid_manager)]

    def _get_resync_interval(self):
        conf = self.grid_manager.grid_config
        try:
            return int(conf.grid_sync_maximum_wait_time)
        except TypeError:
            LOG.warning(_LW("Invalid resync interval set: %s"),
                        conf.grid_sync_maximum_wait_time)
            return self.RESYNC_TRY_INTERVAL

    def _init_periodic_resync(self):
        self.resync_thread = loopingcall.FixedIntervalLoopingCall(
            self._periodic_resync)
        self.resync_thread.start(interval=self.RESYNC_TRY_INTERVAL)

    def _periodic_resync(self):
        try:
            interval = self._get_resync_interval()
            if self.grid_syncer.is_sync_needed(interval):
                LOG.info(_LI("Initiating resync."))
                self.grid_syncer.sync(True)
        except Exception as e:
            LOG.exception(_LE("Resync failed due to error: %s"), e)

    def _init_agent_report_thread(self):
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)
        self.agent_state = {
            'binary': const.AGENT_BINARY_NAME,
            'host': config.CONF.host,
            'topic': 'N/A',
            'configurations': {},
            'start_flag': True,
            'agent_type': const.AGENT_TYPE_INFOBLOX_IPAM}
        self.use_call = True
        if self.report_interval:
            self.report_thread = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            self.report_thread.start(interval=self.report_interval)

    def _report_state(self):
        try:
            self.state_rpc.report_state(self.context, self.agent_state,
                                        self.use_call)
            self.agent_state.pop('start_flag', None)
            self.use_call = False
        except AttributeError:
            # This means the server does not support report_state
            LOG.warning(_LW("infoblox-ipam-agent does not support state "
                            "report. State report for this agent will be "
                            "disabled."))
            self.report_thread.stop()
            return
        except Exception:
            LOG.exception(_LE("Failed reporting state!"))
            return

    def start(self):
        super(NotificationService, self).start()
        self.event_listener = get_notification_listener(
            self.transport,
            self.event_targets,
            self.event_endpoints,
            pool=const.AGENT_NOTIFICATION_POOL,
            executor=self.executor
        )
        self.event_listener.start()

    def stop(self, graceful=False):
        if self.event_listener:
            self.event_listener.stop()
            self.event_listener.wait()
        if self.report_thread:
            self.report_thread.stop()
        super(NotificationService, self).stop(graceful)


def get_notification_listener(transport, targets, endpoints,
                              allow_requeue=False, pool=None,
                              executor='blocking'):
    """Return a configured oslo_messaging notification listener."""
    return oslo_messaging.get_notification_listener(
        transport, targets, endpoints, executor=executor,
        allow_requeue=allow_requeue, pool=pool)



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import decimal
import hashlib
import netaddr
import re
import six
import time
import urllib

from oslo_serialization import jsonutils

from infoblox_client import connector as conn
from infoblox_client import feature
from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import config as cfg
from networking_infoblox.neutron.common import constants as const


def json_to_obj(obj_type, json_data):
    """Converts json data to an object with a given object type

    :param obj_type: converted object's type that is determined dynamically
    :param json_data: json string or json object
    :return: object
    """
    def dic2obj(x):
        if isinstance(x, dict):
            return type(obj_type, (),
                        {k: dic2obj(v) for k, v in six.iteritems(x)})
        else:
            return x

    if isinstance(json_data, six.string_types):
        json_data = jsonutils.loads(json_data)
    return dic2obj(json_data)


def get_values_from_records(key, records):
    key_vals = []
    if records is None:
        return key_vals

    for record in records:
        key_val = record.get(key, None)
        if key_val:
            key_vals.append(key_val)
    return key_vals


def get_composite_values_from_records(keys, records, delimiter='^'):
    key_vals = []
    if records is None:
        return key_vals

    for record in records:
        values = []
        for key in keys:
            key_val = record.get(key)
            if key_val:
                if not isinstance(key_val, six.string_types):
                    key_val = str(key_val)
                values.append(key_val)
        key_vals.append(delimiter.join(values))
    return key_vals


def db_records_to_obj(obj_type, records):
    record_json = db_records_to_json(records)
    if not isinstance(record_json, list):
        return json_to_obj(obj_type, record_json)

    result_set = []
    for record in records:
        result_set.append(json_to_obj(obj_type, record))
    return result_set


def db_records_to_json(records):
    """Converts db records to json.

    alchemy_encoder is needed for date and numeric(x,y) fields since
    they will turn into datetime.date and decimal.Decimal types.
    """
    def alchemy_encoder(obj):
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, decimal.Decimal):
            return float(obj)

    rows = []
    for record in records:
        if isinstance(record, tuple):
            merge = dict()
            for table in record:
                merge.update(dict(table))
            rows.append(merge)
        else:
            rows.append(dict(record))

    # return all rows as a JSON array of objects
    json_str = jsonutils.dumps(rows, alchemy_encoder)
    return jsonutils.loads(json_str)


def get_string_or_none(value):
    ret_val = None
    if isinstance(value, six.string_types):
        ret_val = value
    else:
        if value is not None:
            ret_val = str(value)
    return ret_val


def get_ea_value(name, extattrs, should_return_list_value=False):
    valid = (name and isinstance(name, six.string_types) and
             extattrs and isinstance(extattrs, dict))
    if not valid:
        return None

    value = None
    if extattrs:
        root = extattrs.get("extattrs")
        if root:
            name_attr = root.get(name)
            if name_attr:
                value = name_attr.get('value')
                if should_return_list_value and not isinstance(value, list):
                    value = [value]
    return value


def reset_required_eas(ib_obj):
    if not ib_obj or not ib_obj.extattrs or not ib_obj.extattrs.to_dict():
        return

    for ea in const.REQUIRED_EA_LIST:
        if ea == const.EA_CLOUD_API_OWNED:
            ib_obj.extattrs.set(ea, 'False')
        else:
            ib_obj.extattrs.set(ea, const.EA_RESET_VALUE)


def get_ip_version(ip_address):
    valid = ip_address and isinstance(ip_address, six.string_types)
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if type(ip_address) is dict:
        ip = ip_address['ip_address']
    else:
        ip = ip_address

    try:
        ip = netaddr.IPAddress(ip)
    except ValueError:
        ip = netaddr.IPNetwork(ip)
    return ip.version


def is_valid_ip(ip):
    try:
        netaddr.IPAddress(ip)
    except netaddr.core.AddrFormatError:
        return False
    return True


def get_list_from_string(data_string, delimiter_list):
    valid = (data_string and
             isinstance(data_string, six.string_types) and
             delimiter_list and
             isinstance(delimiter_list, list))
    if not valid:
        raise ValueError("Invalid argument was passed.")

    list_data = remove_any_space(data_string)
    if isinstance(delimiter_list, six.string_types):
        if len(delimiter_list) == 0:
            return data_string
        return list_data.split(delimiter_list)

    if isinstance(delimiter_list, list):
        delimiter_count = len(delimiter_list)
        if delimiter_count == 0:
            return data_string
        if delimiter_count == 1:
            return list_data.split(delimiter_list[0])
        if delimiter_count > 2:
            return ValueError("Delimiter list can contain up to 2 delimiters.")

        result_list = []
        for delimiter in delimiter_list:
            if isinstance(list_data, six.string_types):
                list_data = list_data.split(delimiter)
            else:
                for ld in list_data:
                    result_list.append(ld.split(delimiter))
        # clean up empty string element ['']
        result_list[0] = [m for m in result_list[0] if m]
        result_list[1] = [m for m in result_list[1] if m]
        return result_list

    raise ValueError("Unsupported delimiter list type.")


def exists_in_sequence(sub_sequence_to_find, full_list_in_sequence):
    valid = (isinstance(sub_sequence_to_find, list) and
             isinstance(full_list_in_sequence, list))
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not sub_sequence_to_find or not full_list_in_sequence:
        return False

    return any(full_list_in_sequence[pos:pos + len(sub_sequence_to_find)] ==
               sub_sequence_to_find for pos in
               range(0,
                     len(full_list_in_sequence) -
                     len(sub_sequence_to_find) + 1))


def exists_in_list(list_to_find, full_list):
    valid = isinstance(list_to_find, list) and isinstance(full_list, list)
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not list_to_find or not full_list:
        return False

    found_list = [m for m in list_to_find if m in full_list]
    return len(found_list) == len(list_to_find)


def find_one_in_list(search_key, search_value, search_list):
    """Find one item that match searching one key and value."""
    valid = (isinstance(search_key, six.string_types) and
             isinstance(search_list, list))
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not search_key or not search_value or not search_list:
        return None

    found_list = [m for m in search_list if m.get(search_key) == search_value]
    return found_list[0] if found_list else None


def find_in_list_by_condition(search_key_value_pairs, search_list):
    """Find one item that match given search key value pairs.

    :param search_key_value_pairs: dictionary that contains search key and
    values
    :param search_list: list of dictionary objects to search
    :return: a single item that matches criteria or None
    """
    valid = (isinstance(search_key_value_pairs, dict) and
             isinstance(search_list, list))
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not search_key_value_pairs or not search_list:
        return None

    results = []
    for m in search_list:
        match_failed = False
        for key in search_key_value_pairs:
            if not m.get(key) == search_key_value_pairs[key]:
                match_failed = True
                break
        if not match_failed:
            results.append(m)
    return results


def find_in_list(search_key, search_values, search_list):
    """Find items that match multiple search values on a single search key.

    :param search_key: a key to search
    :param search_values: values that match
    :param search_list: list of dictionary objects to search
    :return: list that matches criteria
    """
    valid = (isinstance(search_key, six.string_types) and
             isinstance(search_values, list) and
             isinstance(search_list, list))
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not search_key or not search_values or not search_list:
        return None

    found_list = [m for m in search_list if m.get(search_key) in search_values]
    return found_list


def find_in_list_by_value(search_value, search_list,
                          first_occurrence_only=True):
    """Find item(s) that match(es) the search value from the search list."""
    valid = isinstance(search_list, list)
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not search_value or not search_list:
        return None

    if isinstance(search_list[0], dict):
        found_list = [m for m in search_list if search_value in m.values()]
    else:
        found_list = [m for m in search_list
                      if search_value in m.__dict__.values()]

    if first_occurrence_only:
        return found_list[0] if found_list else None
    return found_list


def find_key_from_list(search_key, search_list):
    """Find items that contain a search key.

    :param search_key: a key to search
    :param search_list: list of dictionary objects to search
    :return: list that matches criteria
    """
    valid = (isinstance(search_key, six.string_types) and
             isinstance(search_list, list))
    if not valid:
        raise ValueError("Invalid argument was passed.")

    if not search_key or not search_list:
        return None

    found_list = [m for m in search_list
                  if isinstance(m, dict) and m.get(search_key)]
    return found_list


def merge_list(*list_args):
    merge_lsit = []
    for lst in list_args:
        if not isinstance(lst, list):
            raise ValueError("Invalid argument was passed.")
        merge_lsit += lst
    return list(set(merge_lsit))


def remove_any_space(text):
    if text:
        return re.sub(r'\s+', '', text)
    return text


def get_hash(text=None):
    if text and isinstance(text, six.string_types):
        text = text.encode('utf-8')
        return hashlib.md5(text).hexdigest()
    return hashlib.md5(str(time.time())).hexdigest()


def get_oid_from_nios_ref(obj_ref):
    if obj_ref and isinstance(obj_ref, six.string_types) and len(obj_ref) > 0:
        match = re.search('\S+\/(\S+):(\S+)', obj_ref)
        if match:
            return match.group(1)
    return None


def get_network_view_id(grid_id, obj_ref):
    if grid_id and obj_ref:
        obj_id = get_oid_from_nios_ref(obj_ref)
        if obj_id:
            return "%s:%s" % (grid_id, obj_id)
    return None


def get_network_info_from_nios_ref(network_ref):
    if network_ref and len(network_ref) > 0:
        match = re.search('(\S+)\/(\S+):(\S+)\/(\d+)\/(\S+)', network_ref)
        if match:
            cidr = match.group(3) + '/' + match.group(4)
            if match.group(1) == 'ipv6network':
                cidr = urllib.unquote(cidr)
            return {'object_id': match.group(2),
                    'network_view': match.group(5),
                    'cidr': cidr}
    return None


def get_member_status(node_status):
    member_status = const.MEMBER_STATUS_OFF
    if node_status == const.MEMBER_NODE_STATUS_FAILED:
        member_status = const.MEMBER_STATUS_OFF
    elif node_status == const.MEMBER_NODE_STATUS_INACTIVE:
        member_status = const.MEMBER_STATUS_OFF
    elif node_status == const.MEMBER_NODE_STATUS_WARNING:
        member_status = const.MEMBER_STATUS_ON
    elif node_status == const.MEMBER_NODE_STATUS_WORKING:
        member_status = const.MEMBER_STATUS_ON
    return member_status


def get_notification_handler_name(event_type):
    service, resource, action, sequence = (None, None, None, None)
    if event_type.count('.') == 2:
        resource, action, sequence = event_type.split('.', 2)
    else:
        service, resource, action, sequence = event_type.split('.', 3)

    event_sequence = 'alert' if sequence == 'start' else 'sync'
    handler_name = "%s_%s_%s" % (action, resource, event_sequence)
    return handler_name


def generate_network_view_name(object_id, object_name=None):
    """Generates Network View name by id and name.

    Truncates generated network view to do not exceed allowed length of 64
    characters for network_view. dns_view can be generated from network_view
    by prepending 'default.', so limit max allowed length for network view
    by 64 - len('default.') = 56.
    """
    if not object_id or not isinstance(object_id, six.string_types):
        raise ValueError("object_id cannot be empty and must a string.")

    if object_name and not isinstance(object_name, six.string_types):
        raise ValueError("object name must be a string.")

    netview_name = ("{}-{}".format(object_name, object_id)
                    if object_name else object_id)

    if len(netview_name) > const.NETVIEW_MAX_LEN:
        return netview_name[:const.NETVIEW_MAX_LEN]
    return netview_name


def get_connector(credentials=None):
    grid_id = cfg.CONF.infoblox.cloud_data_center_id
    grid_opts = cfg.get_infoblox_grid_opts(grid_id)
    # map connector options to config
    # None as value means no name change needed
    mapping = {'host': 'grid_master_host',
               'username': 'admin_user_name',
               'password': 'admin_password',
               'wapi_version': None,
               'ssl_verify': None,
               'http_pool_connections': None,
               'http_pool_maxsize': None,
               'http_request_timeout': None}
    opts = {field: grid_opts[mapping[field]]
            if mapping[field] else grid_opts[field]
            for field in mapping}
    if opts['ssl_verify'] == 'False':
        opts['silent_ssl_warnings'] = True
    if credentials:
        opts['username'] = credentials['username']
        opts['password'] = credentials['password']
    return conn.Connector(opts)


def get_ipv4_network_prefix(cidr, subnet_name):
    """Add prefix for an ipv4 classless network mask greater than 24."""
    valid = cidr and isinstance(cidr, six.string_types)
    if not valid:
        raise ValueError("Invalid argument was passed.")

    try:
        ip_net = netaddr.IPNetwork(cidr)
    except netaddr.core.AddrFormatError:
        raise ValueError("Invalid cidr")

    if ip_net.version != 4:
        return None

    prefix = None
    if ip_net.prefixlen > 24:
        if subnet_name and len(subnet_name) > 0:
            prefix = subnet_name
        else:
            prefix = '-'.join(filter(None, re.split(r'[.:/]', cidr)))
    return prefix


def get_features(version, feature_versions=None):
    if feature_versions is None:
        feature_versions = const.FEATURE_VERSIONS
    return feature.Feature(version, feature_versions=feature_versions)


def get_dhcp_member_ips(ib_network):
    """Get dhcp member ips from network json or ib network object."""
    member_ips = []
    if (not ib_network or not (isinstance(ib_network, dict) or
                               isinstance(ib_network, ib_objects.Network))):
        return member_ips

    if isinstance(ib_network, dict):
        if ib_network.get('members'):
            for member in ib_network['members']:
                if member.get('_struct') == 'dhcpmember':
                    member_ip = (member.get('ipv4addr') or
                                 member.get('ipv6addr'))
                    if member_ip:
                        member_ips.append(member_ip)
    else:
        if ib_network.members:
            for member in ib_network.members:
                if member._struct == 'dhcpmember':
                    member_ip = member.ipv4addr or member.ipv6addr
                    if member_ip:
                        member_ips.append(member_ip)
    return member_ips


def get_dns_member_ips(ib_network):
    """Get dns member ips from network json or ib network object."""
    member_ips = []
    if (not ib_network or not (isinstance(ib_network, dict) or
                               isinstance(ib_network, ib_objects.Network))):
        return member_ips

    if isinstance(ib_network, dict):
        if ib_network.get('options'):
            for option in ib_network['options']:
                if option.get('name') == 'domain-name-servers':
                    option_values = option.get('value')
                    if option_values:
                        member_ips = option_values.split(',')
                    break
    else:
        for option in ib_network.options:
            if option.name == 'domain-name-servers':
                if option.value:
                    member_ips = option.value.split(',')
                break
    return member_ips


def get_router_ips(ib_network):
    """Get gateway ips (routers) from network json or ib network object."""
    router_ips = []
    if (not ib_network or not (isinstance(ib_network, dict) or
                               isinstance(ib_network, ib_objects.Network))):
        return router_ips

    if isinstance(ib_network, dict):
        if ib_network.get('options'):
            for option in ib_network['options']:
                if option.get('name') == 'routers':
                    option_values = option.get('value')
                    if option_values:
                        router_ips = option_values.split(',')
                    break
    else:
        for option in ib_network.options:
            if option.name == 'routers':
                if option.value:
                    router_ips = option.value.split(',')
                break
    return router_ips


def find_member_by_ip_from_list(member_ip, members):
    """Find a member by ip which could be either ipv4 or ip6."""
    ip_ver = get_ip_version(member_ip)
    if ip_ver == 4:
        member = find_one_in_list('member_ip', member_ip, members)
    else:
        member = find_one_in_list('member_ipv6', member_ip, members)
    return member


def get_nameservers(ib_dns_members, ip_version):
    if (not isinstance(ib_dns_members, list) or
            ip_version not in [4, 6]):
        raise ValueError("Invalid argument was passed.")

    # Prefer member_dns_ipX and fallback to member_ipX if dns one not set
    nameservers = [n for n in [m.member_dns_ipv6 or m.member_ipv6
                               if ip_version == 6
                               else m.member_dns_ip or m.member_ip
                               for m in ib_dns_members] if n]
    return nameservers


def get_mapping_relation(member_type):
    if member_type == const.MEMBER_TYPE_CP_MEMBER:
        return const.MAPPING_RELATION_DELEGATED
    if member_type == const.MEMBER_TYPE_GRID_MASTER:
        return const.MAPPING_RELATION_GM_OWNED
    return None



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import constants as n_const

from infoblox_client import objects as ib_objects

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import ea_manager
from networking_infoblox.tests import base


class EaManagerTestCase(base.TestCase):

    def setUp(self):
        super(EaManagerTestCase, self).setUp()
        self.grid_id = 100
        self.user_id = mock.Mock()
        self.tenant_id = 'Tenant ID'
        self.tenant_name = 'Tenant Name'
        self.context = mock.Mock()
        self.context.tenant_id = self.tenant_id
        self.context.tenant_name = self.tenant_name

    def test_get_common_ea_for_network(self):
        network = {'router:external': True,
                   'shared': True}
        expected_ea = {'Is External': 'True',
                       'Is Shared': 'True',
                       'CMP Type': 'OpenStack',
                       'Cloud API Owned': 'False',
                       'Tenant ID': self.tenant_id,
                       'Tenant Name': self.tenant_name,
                       'Account': self.user_id}
        generated_ea = ea_manager.get_common_ea(network,
                                                self.user_id,
                                                self.tenant_id,
                                                self.tenant_name,
                                                for_network=True)
        self.assertEqual(expected_ea, generated_ea)

    def test_get_common_ea(self):
        network = {'router:external': 'True',
                   'shared': 'True'}
        expected_ea = {'CMP Type': 'OpenStack',
                       'Cloud API Owned': 'False',
                       'Tenant ID': self.tenant_id,
                       'Tenant Name': self.tenant_name,
                       'Account': self.user_id}
        generated_ea = ea_manager.get_common_ea(network,
                                                self.user_id,
                                                self.tenant_id,
                                                self.tenant_name)
        self.assertEqual(expected_ea, generated_ea)

    def test_get_common_ea_no_network(self):
        expected_ea = {'Is External': 'False',
                       'Is Shared': 'False',
                       'CMP Type': 'OpenStack',
                       'Cloud API Owned': 'True',
                       'Tenant ID': self.tenant_id,
                       'Tenant Name': self.tenant_name,
                       'Account': self.user_id}
        generated_ea = ea_manager.get_common_ea(None,
                                                self.user_id,
                                                self.tenant_id,
                                                self.tenant_name,
                                                for_network=True)
        self.assertEqual(expected_ea, generated_ea)

    def test_get_common_ea_cloud_api_owned_true(self):
        owned_true = ({'router:external': False,
                       'shared': False},
                      None,
                      {})
        for network in owned_true:
            generated_ea = ea_manager.get_common_ea(network, None, None, None)
            self.assertEqual(str(True), generated_ea['Cloud API Owned'])

    def test_get_common_ea_cloud_api_owned_false(self):
        owned_false = ({'router:external': True,
                        'shared': False},
                       {'router:external': False,
                        'shared': True},
                       {'router:external': True,
                        'shared': True})
        for network in owned_false:
            generated_ea = ea_manager.get_common_ea(network, None, None, None)
            self.assertEqual(str(False), generated_ea['Cloud API Owned'])

    def test_get_ea_for_network_view(self):
        ea = ea_manager.get_ea_for_network_view(self.tenant_id,
                                                self.tenant_name,
                                                self.grid_id)
        self.assertEqual(self.tenant_id, ea.get('Tenant ID'))
        self.assertEqual(self.tenant_name, ea.get('Tenant Name'))
        self.assertEqual(str(False), ea.get('Cloud API Owned'))
        self.assertEqual('100', ea.get('Cloud Adapter ID'))
        self.assertEqual('OpenStack', ea.get('CMP Type'))

    def test_get_ea_for_network(self):
        network = {'id': mock.Mock(),
                   'name': mock.Mock(),
                   'provider:network_type': mock.Mock(),
                   'provider:physical_network': mock.Mock(),
                   'provider:segmentation_id': mock.Mock(),
                   'router:external': True,
                   'shared': True}
        subnet = {'id': mock.Mock(),
                  'name': mock.Mock()}
        mapping = {'Subnet ID': subnet['id'],
                   'Subnet Name': subnet['name'],
                   'Network ID': network['id'],
                   'Network Name': network['name'],
                   'Network Encap': network['provider:network_type'],
                   'Segmentation ID': network['provider:segmentation_id'],
                   'Physical Network Name': (
                       network['provider:physical_network']),
                   'Tenant ID': self.tenant_id,
                   'Tenant Name': self.tenant_name,
                   'Account': self.user_id,
                   'Is External': str(True),
                   'Is Shared': str(True),
                   'Cloud API Owned': str(False)}
        ea = ea_manager.get_ea_for_network(self.user_id, self.tenant_id,
                                           self.tenant_name, network, subnet)
        for key, value in mapping.items():
            self.assertEqual(mapping[key], ea.get(key))

    def test_get_ea_for_range(self):
        network = {'router:external': False,
                   'shared': False}
        ea = ea_manager.get_ea_for_range(self.user_id, self.tenant_id,
                                         self.tenant_name, network)
        self.assertIsInstance(ea, ib_objects.EA)
        self.assertEqual(str(True), ea.get('Cloud API Owned'))
        self.assertEqual(self.tenant_id, ea.get('Tenant ID'))
        self.assertEqual(self.tenant_name, ea.get('Tenant Name'))
        self.assertEqual(self.user_id, ea.get('Account'))

    def test_get_default_ea_for_ip(self):
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'Port ID': None,
                       'Port Attached Device - Device Owner': None,
                       'Port Attached Device - Device ID': None,
                       'Cloud API Owned': str(True),
                       'IP Type': 'Fixed',
                       'VM ID': None}
        ea = ea_manager.get_default_ea_for_ip(self.user_id, self.tenant_id,
                                              self.tenant_name)
        for key, value in expected_ea.items():
            self.assertEqual(value, ea.get(key))

    def test_get_ea_for_ip(self):
        network = {'router:external': False,
                   'shared': False}
        port_id = mock.Mock()
        device_id = mock.Mock()
        device_owner = mock.Mock()
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'Port ID': port_id,
                       'Port Attached Device - Device Owner': device_owner,
                       'Port Attached Device - Device ID': device_id,
                       'Cloud API Owned': str(True),
                       'IP Type': 'Fixed',
                       'VM ID': None}

        ea = ea_manager.get_ea_for_ip(self.user_id, self.tenant_id,
                                      self.tenant_name, network,
                                      port_id, device_id, device_owner)
        for key, value in expected_ea.items():
            self.assertEqual(value, ea.get(key))

    def test_get_ea_for_ip_with_router_gateway_ip(self):
        network = {'router:external': False,
                   'shared': False}
        port_id = mock.Mock()
        device_id = mock.Mock()
        device_owner = n_const.DEVICE_OWNER_ROUTER_GW
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'Port ID': port_id,
                       'Port Attached Device - Device Owner': device_owner,
                       'Port Attached Device - Device ID': device_id,
                       'Cloud API Owned': str(True),
                       'IP Type': 'Fixed',
                       'VM ID': None}

        ea = ea_manager.get_ea_for_ip(self.user_id, self.tenant_id,
                                      self.tenant_name, network,
                                      port_id, device_id, device_owner)
        for key, value in expected_ea.items():
            self.assertEqual(value, ea.get(key))

    def test_get_ea_for_ip_with_floatingip_creation(self):
        network = {'router:external': True,
                   'shared': False}
        port_id = mock.Mock()
        device_id = mock.Mock()
        device_owner = n_const.DEVICE_OWNER_FLOATINGIP
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'Port ID': port_id,
                       'Port Attached Device - Device Owner': device_owner,
                       'Port Attached Device - Device ID': device_id,
                       'Cloud API Owned': str(False),
                       'IP Type': 'Floating',
                       'VM ID': None}

        ea = ea_manager.get_ea_for_ip(self.user_id, self.tenant_id,
                                      self.tenant_name, network, port_id,
                                      device_id, device_owner)
        for key, value in expected_ea.items():
            self.assertEqual(value, ea.get(key))

    def test_get_ea_for_ip_with_floatingip_association(self):
        network = {'router:external': True,
                   'shared': False}
        port_id = mock.Mock()
        device_id = mock.Mock()
        device_owner = const.NEUTRON_DEVICE_OWNER_COMPUTE_NOVA
        is_floating_ip = True
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'Port ID': port_id,
                       'Port Attached Device - Device Owner': device_owner,
                       'Port Attached Device - Device ID': device_id,
                       'Cloud API Owned': str(False),
                       'IP Type': 'Floating',
                       'VM ID': device_id}

        ea = ea_manager.get_ea_for_ip(self.user_id, self.tenant_id,
                                      self.tenant_name,
                                      network, port_id, device_id,
                                      device_owner, is_floating_ip)
        for key, value in expected_ea.items():
            self.assertEqual(value, ea.get(key))

    def test_get_ea_for_ip_with_floatingip_dissociation(self):
        network = {'router:external': True,
                   'shared': False}
        port_id = mock.Mock()
        device_id = mock.Mock()
        device_owner = n_const.DEVICE_OWNER_FLOATINGIP
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'Port ID': port_id,
                       'Port Attached Device - Device Owner': device_owner,
                       'Port Attached Device - Device ID': device_id,
                       'Cloud API Owned': str(False),
                       'IP Type': 'Floating',
                       'VM ID': None}

        ea = ea_manager.get_ea_for_ip(self.user_id, self.tenant_id,
                                      self.tenant_name,
                                      network, port_id, device_id,
                                      device_owner)
        for key, value in expected_ea.items():
            self.assertEqual(value, ea.get(key))

    def test_get_ea_for_zones_without_net(self):
        # Test that None values is acceptable for network/subnet parameters
        expected_ea = {'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'CMP Type': 'OpenStack',
                       'Cloud API Owned': str(True)}
        ea = ea_manager.get_ea_for_reverse_zone(self.user_id, self.tenant_id,
                                                self.tenant_name, None, None)
        self.assertEqual(expected_ea, ea.ea_dict)
        ea = ea_manager.get_ea_for_forward_zone(self.user_id, self.tenant_id,
                                                self.tenant_name, None, None,
                                                '{subnet_id}.infoblox.com')
        self.assertEqual(expected_ea, ea.ea_dict)

    def test_get_ea_for_reverse_zone(self):
        network = {'id': mock.Mock(),
                   'name': mock.Mock(),
                   'provider:segmentation_id': mock.Mock(),
                   'provider:physical_network': mock.Mock(),
                   'provider:network_type': mock.Mock()}
        subnet = {'id': mock.Mock(),
                  'name': mock.Mock()}
        expected_ea = {'Subnet ID': subnet['id'],
                       'Subnet Name': subnet['name'],
                       'Network ID': network['id'],
                       'Network Name': network['name'],
                       'Network Encap': network['provider:network_type'],
                       'Segmentation ID': network['provider:segmentation_id'],
                       'Physical Network Name': (
                           network['provider:physical_network']),
                       'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'CMP Type': 'OpenStack',
                       'Cloud API Owned': str(True)}
        ea = ea_manager.get_ea_for_reverse_zone(self.user_id, self.tenant_id,
                                                self.tenant_name, network,
                                                subnet)
        self.assertEqual(expected_ea, ea.ea_dict)

    def _test_get_ea_for_forward_zone(self, template, skip_eas):
        network = {'id': mock.Mock(),
                   'name': mock.Mock(),
                   'provider:segmentation_id': mock.Mock(),
                   'provider:physical_network': mock.Mock(),
                   'provider:network_type': mock.Mock()}
        subnet = {'id': mock.Mock(),
                  'name': mock.Mock()}
        verified_ea = {'Subnet ID': subnet['id'],
                       'Subnet Name': subnet['name'],
                       'Network ID': network['id'],
                       'Network Name': network['name'],
                       'Network Encap': network['provider:network_type'],
                       'Segmentation ID': network['provider:segmentation_id'],
                       'Physical Network Name': (
                           network['provider:physical_network']),
                       'Tenant ID': self.tenant_id,
                       'Tenant Name': self.context.tenant_name,
                       'Account': self.user_id,
                       'CMP Type': 'OpenStack',
                       'Cloud API Owned': str(True)}
        ea = ea_manager.get_ea_for_forward_zone(self.user_id, self.tenant_id,
                                                self.tenant_name, network,
                                                subnet, template)
        expected_ea = {key: value for key, value in verified_ea.items()
                       if key not in skip_eas}
        self.assertEqual(expected_ea, ea.ea_dict)

    def test_get_ea_for_forward_zone_template_subnet(self):
        self._test_get_ea_for_forward_zone(template='{subnet_id}',
                                           skip_eas=())
        self._test_get_ea_for_forward_zone(template='{subnet_name}',
                                           skip_eas=())

    def test_get_ea_for_forward_zone_template_network(self):
        skip_eas = ('Subnet ID', 'Subnet Name')
        self._test_get_ea_for_forward_zone(template='{network_id}',
                                           skip_eas=skip_eas)
        self._test_get_ea_for_forward_zone(template='{network_name}',
                                           skip_eas=skip_eas)

    def test_get_ea_for_forward_zone_no_net_in_template(self):
        skip_eas = ('Subnet ID', 'Subnet Name', 'Network ID', 'Network Name',
                    'Network Encap', 'Segmentation ID',
                    'Physical Network Name')
        self._test_get_ea_for_forward_zone(
            template='private.infoblox.com', skip_eas=skip_eas)

    def test_reset_ea_for_network(self):
        network_ea = {'CMP Type': {'value': 'OpenStack'},
                      'Cloud API Owned': {'value': 'True'},
                      'Tenant ID': {'value': 'test-id'},
                      'Tenant Name': {'value': 'tenant-name'},
                      'Account': {'value': 'admin'},
                      'Is External': {'value': 'False'},
                      'Is Shared': {'value': 'True'},
                      'Network ID': {'value': 'True'},
                      'Network Name': {'value': 'True'},
                      'Subnet ID': {'value': 'True'},
                      'Subnet Name': {'value': 'True'},
                      'Network Encap': {'value': 'gre'},
                      'Segmentation ID': {'value': 'segmentation-id'},
                      'Physical Network Name': {'value': 'physical-network'}}
        ib_network_ea = ib_objects.EA.from_dict(network_ea)
        ib_network_mock = mock.Mock(extattrs=ib_network_ea)

        ea_manager.reset_ea_for_network(ib_network_mock)

        expected_ea = {'Cloud API Owned': {'value': 'N/A'},
                       'CMP Type': {'value': 'N/A'},
                       'Tenant ID': {'value': 'N/A'}}
        ib_network_mock.extattrs.ea_dict == expected_ea

    def test_reset_ea_for_range(self):
        range_ea = {'CMP Type': {'value': 'OpenStack'},
                    'Cloud API Owned': {'value': 'True'},
                    'Tenant ID': {'value': 'test-id'},
                    'Tenant Name': {'value': 'tenant-name'},
                    'Account': {'value': 'admin'}}
        ib_range_ea = ib_objects.EA.from_dict(range_ea)
        ib_range_mock = mock.Mock(extattrs=ib_range_ea)

        ea_manager.reset_ea_for_range(ib_range_mock)

        expected_ea = {'Cloud API Owned': {'value': 'N/A'},
                       'CMP Type': {'value': 'N/A'},
                       'Tenant ID': {'value': 'N/A'}}
        ib_range_mock.extattrs.ea_dict == expected_ea

    def test_reset_ea_for_zone(self):
        zone_ea = {'CMP Type': {'value': 'OpenStack'},
                   'Cloud API Owned': {'value': 'True'},
                   'Tenant ID': {'value': 'test-id'},
                   'Tenant Name': {'value': 'tenant-name'},
                   'Account': {'value': 'admin'}}
        ib_zone_ea = ib_objects.EA.from_dict(zone_ea)
        ib_zone_mock = mock.Mock(extattrs=ib_zone_ea)

        ea_manager.reset_ea_for_zone(ib_zone_mock)

        expected_ea = {'Cloud API Owned': {'value': 'N/A'},
                       'CMP Type': {'value': 'N/A'},
                       'Tenant ID': {'value': 'N/A'}}
        ib_zone_mock.extattrs.ea_dict == expected_ea



# Copyright (c) 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests.unit import testlib_api
from neutron_lib import context

from networking_infoblox.neutron.common import constants as const
from networking_infoblox.neutron.common import grid
from networking_infoblox.neutron.common import mapping
from networking_infoblox.neutron.common import member
from networking_infoblox.neutron.common import utils
from networking_infoblox.neutron.db import infoblox_db as dbi

from networking_infoblox.tests import base


DELIMITER = '^'


class GridMappingTestCase(base.TestCase, testlib_api.SqlTestCase):

    def setUp(self):

        super(GridMappingTestCase, self).setUp()
        self.ctx = context.get_admin_context()

        self.test_grid_config = grid.GridConfiguration(self.ctx)
        self.test_grid_config.gm_connector = mock.Mock()
        self.test_grid_config.grid_id = 100
        self.test_grid_config.grid_name = "Test Grid 1"
        self.test_grid_config.grid_master_host = '192.168.1.7'
        self.test_grid_config.grid_master_name = 'nios-7.2.0-master.com'
        self.test_grid_config.admin_user_name = 'admin'
        self.test_grid_config.admin_password = 'infoblox'
        self.test_grid_config.wapi_version = '2.2'
        self.member_mgr = member.GridMemberManager(self.test_grid_config)

    def _create_members_with_cloud(self):
        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITH_CLOUD)
        license_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBER_LICENSES)

        self.member_mgr._discover_members = mock.Mock(return_value=member_json)
        self.member_mgr._discover_member_licenses = mock.Mock(
            return_value=license_json)

        self.member_mgr._discover_dns_settings = mock.Mock(return_value=[])
        self.member_mgr._discover_dhcp_settings = mock.Mock(return_value=[])

        self.member_mgr.sync()

    def _create_members_without_cloud(self):
        member_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_MEMBERS_WITHOUT_CLOUD)
        self.member_mgr._discover_members = mock.Mock()
        self.member_mgr._discover_members.return_value = member_json

        self.member_mgr._discover_member_licenses = mock.Mock()
        self.member_mgr._discover_member_licenses.return_value = None

        self.member_mgr.sync()

    def _validate_network_views(self, network_view_json):
        db_network_views = dbi.get_network_views(self.ctx.session)
        self.assertEqual(len(network_view_json), len(db_network_views))
        expected = [nv['name'] for nv in network_view_json]
        actual = utils.get_values_from_records('network_view',
                                               db_network_views)
        self.assertEqual(expected, actual)

    def _validate_mapping_conditions(self, network_view_json):
        db_network_views = dbi.get_network_views(self.ctx.session)
        db_mapping_conditions = dbi.get_mapping_conditions(self.ctx.session)

        expected_conditions = dict((nv['name'], nv['extattrs'])
                                   for nv in network_view_json
                                   if nv['extattrs'])
        expected_condition_rows = []
        for netview in expected_conditions:
            netview_row = utils.find_one_in_list('network_view', netview,
                                                 db_network_views)
            netview_id = netview_row.id
            for condition_name in expected_conditions[netview]:
                if 'Mapping' not in condition_name:
                    continue
                values = expected_conditions[netview][condition_name]['value']
                if not isinstance(values, list):
                    expected_condition_rows.append(netview_id + DELIMITER +
                                                   condition_name + DELIMITER +
                                                   values)
                    continue
                for value in values:
                    expected_condition_rows.append(netview_id + DELIMITER +
                                                   condition_name + DELIMITER +
                                                   value)

        actual_condition_rows = utils.get_composite_values_from_records(
            ['network_view_id', 'neutron_object_name', 'neutron_object_value'],
            db_mapping_conditions)
        self.assertEqual(set(expected_condition_rows),
                         set(actual_condition_rows))

    def _validate_member_mapping(self, network_view_json, network_json):
        db_members = dbi.get_members(self.ctx.session,
                                     grid_id=self.test_grid_config.grid_id)
        db_network_views = dbi.get_network_views(self.ctx.session)
        db_mapping_members = dbi.get_mapping_members(self.ctx.session)
        db_service_members = dbi.get_service_members(self.ctx.session)

        gm_row = utils.find_one_in_list('member_type',
                                        const.MEMBER_TYPE_GRID_MASTER,
                                        db_members)
        gm_member_id = gm_row.member_id

        dedicated_delegation_members = dict()
        for netview in network_view_json:
            netview_name = netview['name']
            if (netview.get('cloud_info') and
                    netview.get('cloud_info').get('delegated_member')):
                delegated_member = utils.find_one_in_list(
                    'member_name',
                    netview['cloud_info']['delegated_member']['name'],
                    db_members)
                dedicated_delegation_members[netview_name] = (
                    delegated_member.member_id)

        expected_mapping_members = []
        expected_service_members = []

        # get delegated authority members from network views
        for netview in dedicated_delegation_members:
            netview_row = utils.find_one_in_list('network_view', netview,
                                                 db_network_views)
            netview_id = netview_row.id
            authority_member = dedicated_delegation_members[netview]
            mapping_relation = const.MAPPING_RELATION_DELEGATED
            mapping_row_info = (netview_id + DELIMITER + authority_member +
                                DELIMITER + mapping_relation)
            expected_mapping_members.append(mapping_row_info)

        # get authority members from networks
        for network in network_json:
            netview = network['network_view']
            netview_row = utils.find_one_in_list('network_view', netview,
                                                 db_network_views)
            netview_id = netview_row.id

            mapping_relation = const.MAPPING_RELATION_GM_OWNED
            authority_member = gm_member_id
            if netview in dedicated_delegation_members:
                authority_member = dedicated_delegation_members[netview]
                mapping_relation = const.MAPPING_RELATION_DELEGATED
            elif (network.get('cloud_info') and
                    network['cloud_info'].get('delegated_member')):
                delegated_member = utils.find_one_in_list(
                    'member_name',
                    network['cloud_info']['delegated_member']['name'],
                    db_members)
                authority_member = delegated_member.member_id
                mapping_relation = const.MAPPING_RELATION_DELEGATED

            mapping_row_info = (netview_id + DELIMITER + authority_member +
                                DELIMITER + mapping_relation)
            if mapping_row_info not in expected_mapping_members:
                expected_mapping_members.append(mapping_row_info)

            if network.get('members'):
                for m in network['members']:
                    if m['_struct'] == 'dhcpmember':
                        dhcp_member = utils.find_one_in_list(
                            'member_name', m['name'], db_members)
                        mapping_row_info = (netview_id + DELIMITER +
                                            dhcp_member.member_id + DELIMITER +
                                            const.SERVICE_TYPE_DHCP)
                        if mapping_row_info not in expected_service_members:
                            expected_service_members.append(mapping_row_info)

            if network.get('options'):
                dns_membe_ips = []
                for option in network['options']:
                    if option.get('name') == 'domain-name-servers':
                        option_values = option.get('value')
                        if option_values:
                            dns_membe_ips = option_values.split(',')
                            break
                for membe_ip in dns_membe_ips:
                    dns_member = utils.find_one_in_list(
                        'member_ip', membe_ip, db_members)
                    mapping_row_info = (netview_id + DELIMITER +
                                        dns_member.member_id + DELIMITER +
                                        const.SERVICE_TYPE_DNS)
                    if mapping_row_info not in expected_service_members:
                        expected_service_members.append(mapping_row_info)

        actual_mapping_members = utils.get_composite_values_from_records(
            ['network_view_id', 'member_id', 'mapping_relation'],
            db_mapping_members)
        self.assertEqual(set(expected_mapping_members),
                         set(actual_mapping_members))

        actual_service_members = utils.get_composite_values_from_records(
            ['network_view_id', 'member_id', 'service'],
            db_service_members)
        self.assertEqual(set(expected_service_members),
                         set(actual_service_members))

    def test_sync_for_cloud(self):
        self._create_members_with_cloud()

        mapping_mgr = mapping.GridMappingManager(self.test_grid_config)
        mapping_mgr._sync_nios_for_network_view = mock.Mock()

        network_view_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_NETWORKVIEW_WITH_CLOUD)
        mapping_mgr._discover_network_views = mock.Mock()
        mapping_mgr._discover_network_views.return_value = network_view_json

        network_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_NETWORK_WITH_CLOUD)
        mapping_mgr._discover_networks = mock.Mock()
        mapping_mgr._discover_networks.return_value = network_json

        dnsview_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_DNS_VIEW)
        mapping_mgr._discover_dns_views = mock.Mock()
        mapping_mgr._discover_dns_views.return_value = dnsview_json

        mapping_mgr.sync()

        # validate network views, mapping conditions, mapping members
        self._validate_network_views(network_view_json)
        self._validate_mapping_conditions(network_view_json)
        self._validate_member_mapping(network_view_json, network_json)

    def test_sync_for_without_cloud(self):
        self._create_members_with_cloud()

        mapping_mgr = mapping.GridMappingManager(self.test_grid_config)
        mapping_mgr._sync_nios_for_network_view = mock.Mock()

        network_view_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_NETWORKVIEW_WITHOUT_CLOUD)
        mapping_mgr._discover_network_views = mock.Mock()
        mapping_mgr._discover_network_views.return_value = network_view_json

        network_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_NETWORK_WITHOUT_CLOUD)
        mapping_mgr._discover_networks = mock.Mock()
        mapping_mgr._discover_networks.return_value = network_json

        dnsview_json = self.connector_fixture.get_object(
            base.FixtureResourceMap.FAKE_DNS_VIEW)
        mapping_mgr._discover_dns_views = mock.Mock()
        mapping_mgr._discover_dns_views.return_value = dnsview_json

        mapping_mgr.sync()

        # validate network views, mapping conditions, mapping members
        self._validate_network_views(network_view_json)
        self._validate_mapping_conditions(network_view_json)
        self._validate_member_mapping(network_view_json, network_json)
